package CPANPLUS::Dist::Sample;

use vars qw[$VERSION];
$VERSION = "0.9916";

=pod

=head1 NAME

CPANPLUS::Dist::Sample -- Sample code to create your own Dist::* plugin

=head1 Description.

This document is B<Obsolete>. Please read the documentation and code
in C<CPANPLUS::Dist::Base>.

=cut

1;
